package de.fitko.fitconnect.client;

import de.fitko.fitconnect.api.domain.auth.OAuthToken;
import de.fitko.fitconnect.api.domain.model.metadata.Metadata;
import de.fitko.fitconnect.api.domain.model.metadata.attachment.Attachment;
import de.fitko.fitconnect.api.domain.model.metadata.data.Data;
import de.fitko.fitconnect.api.domain.model.submission.Submission;
import de.fitko.fitconnect.api.exceptions.client.AttachmentUploadError;
import de.fitko.fitconnect.api.exceptions.client.ClientNotAuthenticated;
import de.fitko.fitconnect.api.services.Sender;
import de.fitko.fitconnect.api.services.metadata.MetadataService;

import java.util.List;
import java.util.Optional;

/**
 * A fluent client for announcing and handing in a {@link Submission}
 */
public class SenderClient extends Client {

    private SenderClient() {}

     public static Authenticate builder() {
         final Sender sender = getService(Sender.class);
         final MetadataService metadataService = getService(MetadataService.class);
         return new ClientBuilder(sender, metadataService);
    }

    public interface Authenticate {
        /**
         * Authenticates the client against the FitConnect-Auth-API and retrieves an {@link OAuthToken}.
         *
         * @param clientId a unique client identifier
         * @param secret the applications secret key
         * @param scope 1 or more client scopes that determine if a submission is accepted by the client
         * @return the step for creating a new submission if the authentication was successful
         *
         * @throws ClientNotAuthenticated if the OAuth token could not be retrieved
         */
        CreateSubmission authenticate(String clientId, String secret, String... scope) throws ClientNotAuthenticated;
    }

    public interface AttachmentUpload {
        /**
         * Encrypts and Uploads all {@link Attachment}s to the FitConnect-API
         *
         * @param attachments a list of attachments that are uploaded
         * @return the step for sending the submission if the creating step was successful
         *
         * @throws ClientNotAuthenticated if the client is not authenticated
         * @throws AttachmentUploadError if the upload failed
         */
        SendSubmission uploadAttachments(List<Attachment> attachments);
    }

    public interface CreateSubmission {
        /**
         * Creates and announces a new {@link Submission}
         *
         * @param submission the new {@link Submission} that should be created on server-side
         * @return the step for upload the attachments if the creation was successful
         */
        AttachmentUpload createSubmission(Submission submission);
    }

    public interface SendSubmission {
        /**
         * Encrypts and uploads  the {@link Metadata} and {@link Data} and hands in the {@link Submission} at
         * the defined destination-id.
         *
         * @param metaData the submissions {@link Metadata}
         * @param data the submissions {@link Data}
         *
         * @return the finally uploaded and sent submission, empty if an error occurred
         */
        Optional<Submission> sendSubmission(Metadata metaData, Data data);
    }

    public static class ClientBuilder implements Authenticate, AttachmentUpload, CreateSubmission, SendSubmission {

        private final MetadataService metadataService;
        private Sender sender;
        private Optional<OAuthToken> token;

        private ClientBuilder(Sender sender, MetadataService metadataService) {
            this.sender = sender;
            this.metadataService = metadataService;
        }

        @Override
        public CreateSubmission authenticate(String clientId, String secret, String... scope) {
            this.token = sender.retrieveOAuthToken(clientId, secret, scope);
            checkIfAuthenticated();
            return this;
        }

        @Override
        public SendSubmission uploadAttachments(List<Attachment> attachments) {
            checkIfAuthenticated();
            return this;
        }

        @Override
        public AttachmentUpload createSubmission(Submission submission) {
            checkIfAuthenticated();
            return this;
        }

        @Override
        public Optional<Submission> sendSubmission(Metadata metaData, Data data) {
            checkIfAuthenticated();
            return sender.sendSubmission(Submission.builder().build());
        }

        private void checkIfAuthenticated(){
            if(this.token == null || this.token.isEmpty()){
                throw new ClientNotAuthenticated("Not authenticated, please authenticate first");
            }
        }
    }
}